"""hmat.py: :class:`HMat`, :func:`build_hmatrix`, :func:`recursion_build_hmatrix`, :class:`StructureWarning`
"""
import math
import numbers
import numpy
import copy

from HierMat.rmat import RMat


class HMat(object):
    """Implement a hierarchical Matrix
    
    :param blocks: list of HMat instances (children) (optional)
    :type blocks: list(HMat)
    :param content: the content if the matrix has no children (optional)
    :type content: RMat or numpy.matrix
    :param shape: the shape of the matrix (same as for numpy matrices)
    :type shape: tuple(int, int)
    :param parent_index: the index of this matrix with respect to its containing *parent*-matrix, zero based
    :type parent_index: tuple(int, int)
    
    .. admonition:: Supported operations
        
        * ``+`` (formatted addition)
        
        * ``*`` (formatted multiplication)
        
        * ``-a`` (unary minus)
        
        * ``-`` (formatted subtraction)
        
        * ``==`` (equal)
        
        * ``!=`` (not equal)
        
        * ``abs`` (frobenius norm)
        
        * ``inv`` (inversion)

    """

    def __init__(self, blocks=(), content=None, shape=(), parent_index=()):
        """Implement a hierarchical Matrix
    
        :param blocks: list of HMat instances the children (optional)
        :type blocks: list(HMat)
        :param content: the content if the matrix has no children (optional)
        :type content: RMat or numpy.matrix
        :param shape: the shape of the matrix (same as for numpy matrices)
        :type shape: tuple(int, int)
        :param parent_index: the index of this matrix with respect to its containing *parent*-matrix, zero based
        :type parent_index: tuple(int, int)
        """
        self.blocks = blocks  # This list contains the lower level HMats
        self.content = content  # If not empty, this is either a full matrix or a RMat
        self.shape = shape  # Tuple of dimensions, i.e. size of index sets
        self.parent_index = parent_index  # Tuple of coordinates for the top-left corner in the root matrix

    def __getitem__(self, item):
        """Get block at position i, j from root-index or ith block from blocks"""
        if isinstance(item, int):
            return self.blocks[item]
        else:
            structured_blocks = {block.parent_index: block for block in self.blocks}
            return structured_blocks[item]

    def __setitem__(self, key, value):
        """Set block at index key to value"""
        if isinstance(key, int):
            self.blocks[key] = value
        else:
            index = self.blocks.index(self[key])
            self.blocks[index] = value

    def __repr__(self):
        return '<HMat with {content}>'.format(content=self.blocks if self.blocks else self.content)

    def __eq__(self, other):
        """test for equality
        
        :param other: other HMat
        :type other: HMat
        :return: true on equal
        :rtype: bool
        """
        if not isinstance(self, type(other)):
            return False
        length = len(self.blocks)
        if len(other.blocks) != length:
            return False
        eqs = [a == b for a in self.blocks for b in other.blocks]
        if sum(eqs) != length:  # ignore order of blocks
            return False
        if not isinstance(self.content, type(other.content)):
            return False
        if isinstance(self.content, RMat) and self.content != other.content:
            return False
        if isinstance(self.content, numpy.matrix) and not numpy.array_equal(self.content, other.content):
            return False
        if self.shape != other.shape:
            return False
        if self.parent_index != other.parent_index:
            return False
        return True

    def __ne__(self, other):
        """test for inequality
        
        :param other: other HMat
        :type other: HMat
        :return: true on not equal
        :rtype: bool
        """
        return not self == other

    def __abs__(self):
        """Frobenius norm"""
        if isinstance(self.content, RMat):
            return self.content.norm()
        elif isinstance(self.content, numpy.matrix):
            return numpy.linalg.norm(self.content)
        else:
            total = 0
            for block in self.blocks:
                total += abs(block)**2
            return math.sqrt(total)

    def norm(self, order=None):
        """Norm of the matrix

        :param order: order of the norm (see in :func:`numpy.linalg.norm`)
        :return: norm
        :rtype: float
        """
        if order is None or order == 'fro':
            return abs(self)
        else:
            raise NotImplementedError('Only Frobenius implemented so far')

    def __add__(self, other):
        """addition with several types"""
        try:
            if self.shape != other.shape:
                raise ValueError("operands could not be broadcast together with shapes"
                                 " {0.shape} {1.shape}".format(self, other))
        except AttributeError:
            if not isinstance(other, numbers.Number):
                raise NotImplementedError('unsupported operand type(s) for +: {0} and {1}'.format(type(self),
                                                                                                  type(other)))
            else:
                return self._add_matrix(numpy.matrix(other))
        if isinstance(other, HMat):
            return self._add_hmat(other)
        elif isinstance(other, RMat):
            return self._add_rmat(other)
        elif isinstance(other, numpy.matrix):
            return self._add_matrix(other)
        else:
            raise NotImplementedError('unsupported operand type(s) for +: {0} and {1}'.format(type(self), type(other)))

    def _add_hmat(self, other):
        """add two hmat objects
        
        if structures do not match refine to finer structure
        
        :param other: HMat to add
        :type other: HMat
        :return: sum
        :rtype: HMat
        """
        # check inputs
        if self.parent_index != other.parent_index:
            raise ValueError('can not add {0} and {1}. root indices {0.parent_index} '
                             'and {1.parent_index} not the same'.format(self, other))
        if self.content is None and other.blocks == ():
            # only other has content, so restructure other to match structure
            addend = other.restructure(self.block_structure())
            return self + addend
        elif self.blocks == () and other.content is None:
            # only self has content, so restructure self to match structure
            addend = self.restructure(other.block_structure())
            return addend + other
        # both have content
        elif isinstance(self.content, numpy.matrix) and isinstance(other.content, RMat):
            # # take other first to avoid numpy broadcast
            # return HMat(content=other.content + self.content, shape=self.shape, parent_index=self.parent_index)
            # take other to full
            return HMat(content=self.content + other.to_matrix(), shape=self.shape, parent_index=self.parent_index)
        elif self.content is not None:  # both have content, that can be added left to right
            return HMat(content=self.content + other.content, shape=self.shape, parent_index=self.parent_index)
        # if we get here, both have children
        elif len(self.blocks) == len(other.blocks):
            blocks = [self[index] + other[index] for index in self.block_structure()]
            return HMat(blocks=blocks, shape=self.shape, parent_index=self.parent_index)
        else:
            raise ValueError('can not add {0} and {1}. number of blocks are different'.format(self, other))

    def _add_rmat(self, other):
        # TODO: What here?
        raise NotImplementedError()

    def _add_matrix(self, other):
        """add full matrix to hmat
        
        :param other: matrix to add
        :type other: numpy.matrix
        :return: sum
        :rtype: HMat
        """
        if self.shape != other.shape:
            raise ValueError('operands could not be broadcast together with shapes'
                             ' {0.shape} {1.shape}'.format(self, other))
        out = HMat(shape=self.shape, parent_index=self.parent_index)
        if self.blocks != ():
            out.blocks = []
            for block in self.blocks:
                start_x = block.parent_index[0]
                start_y = block.parent_index[1]
                end_x = start_x + block.shape[0]
                end_y = start_y + block.shape[1]
                out.blocks.append(block + other[start_x: end_x, start_y: end_y])
        else:
            out.content = self.content + other
        return out

    def __radd__(self, other):
        """Should be commutative so just switch"""
        return self + other

    def __neg__(self):
        """Unary minus"""
        out = HMat(shape=self.shape, parent_index=self.parent_index)
        if self.content is not None:
            out.content = -self.content
        else:
            out.blocks = []
            for block in self.blocks:
                out.blocks.append(-block)
        return out

    def __sub__(self, other):
        """Subtract other
        
        :type other: HMat
        """
        return self + (-other)

    def __mul__(self, other):
        """multiplication with several types"""
        if isinstance(other, numpy.matrix):  # order is important here!
            return self._mul_with_matrix(other)
        elif isinstance(other, numpy.ndarray):
            return self._mul_with_vector(other)
        elif isinstance(other, numbers.Number):
            return self._mul_with_scalar(other)
        elif isinstance(other, RMat):
            return self._mul_with_rmat(other)
        elif isinstance(other, HMat):
            return self._mul_with_hmat(other)
        else:
            raise NotImplementedError('unsupported operand type(s) for *: {0} and {1}'.format(type(self), type(other)))

    def __rmul__(self, other):
        """multiply from the right
        
        only supported for scalars
        enables to write ``a * H``
        """
        if not isinstance(other, numbers.Number):
            raise NotImplementedError('unsupported operand type(s) for *: {0} and {1}'.format(type(self), type(other)))
        return self * other

    def _mul_with_vector(self, other):
        """multiply with a vector

        :param other: vector
        :type other: numpy.array
        :return: result vector
        :rtype: numpy.array
        """
        if self.content is not None:
            # We have content. Multiply and return
            return self.content * other
        else:
            if self.shape[1] != other.shape[0]:
                raise ValueError('shapes {0.shape} and {1.shape} not aligned: '
                                 '{0.shape[1]} (dim 1) != {1.shape[0]} (dim 0)'.format(self, other))
            res_length = self.shape[0]
            res = numpy.zeros((res_length, 1))
            for block in self.blocks:
                res_index_start, in_index_start = block.parent_index
                x_length, y_length = block.shape
                in_index_end = in_index_start + y_length
                res_index_end = res_index_start + x_length
                res[res_index_start: res_index_end] += block * other[in_index_start: in_index_end]
            return res

    def _mul_with_matrix(self, other):
        """multiply with a numpy.matrix

        :param other: matrix
        :type other: numpy.matrix
        :return: result matrix
        :rtype: numpy.matrix
        """
        # TODO: should this be a HMat?
        if self.content is not None:
            return self.content * other
        else:
            if self.shape[1] != other.shape[0]:
                raise ValueError('shapes {0.shape} and {1.shape} not aligned: '
                                 '{0.shape[1]} (dim 1) != {1.shape[0]} (dim 0)'.format(self, other))
            res_shape = (self.shape[0], other.shape[1])
            res = numpy.zeros(res_shape)
            for block in self.blocks:
                row_count, col_count = block.shape
                row_start, col_current = block.parent_index
                row_end = row_start + row_count
                res[row_start:row_end, :] += block * other[row_start:row_end, :]
            return res

    def _mul_with_rmat(self, other):
        """multiplication with an rmat

        :param other: rmat to multiply
        :type other: RMat
        :return: Hmat containing the product
        :rtype: HMat
        """
        out = HMat(shape=(self.shape[0], other.shape[1]), parent_index=self.parent_index)
        if isinstance(self.content, RMat):
            out.content = self.content * other
        elif self.content is not None:
            out.content = other.__rmul__(self.content)
        else:
            # TODO: Check this for correctness
            raise TypeError("Encountered HMat with blocks * RMat! What should I do?!")
        return out

    def _mul_with_hmat(self, other):
        """multiplication with other hmat
        
        :type other: HMat
        """
        if self.shape[1] != other.shape[0]:
            raise ValueError('shapes {0.shape} and {1.shape} not aligned: '
                             '{0.shape[1]} (dim 1) != {1.shape[0]} (dim 0)'.format(self, other))
        out_shape = (self.shape[0], other.shape[1])
        if self.parent_index[1] != other.parent_index[0]:
            raise ValueError('root indices {0.parent_index} and {1.parent_index} not aligned: '
                             '{0.parent_index[1]} (dim 1) != {1.parent_index[0]} (dim 0)'.format(self, other))
        out_parent_index = (self.parent_index[0], other.parent_index[1])
        if self.content is not None and other.content is not None:  # simplest case, both have content
            out_content = self.content * other.content
            return HMat(content=out_content, shape=out_shape, parent_index=out_parent_index)
        elif self.content is None and other.content is None:  # both have blocks
            return self._mul_with_hmat_blocks(other, out_shape, out_parent_index)
        elif self.content is not None:  # other has blocks, self has content. Collect other to full matrix
            return HMat(content=self.content * other.to_matrix(), shape=out_shape, parent_index=out_parent_index)
        else:  # other has content, self has blocks. Collect self to full
            return HMat(content=self.to_matrix() * other.content, shape=out_shape, parent_index=out_parent_index)

    def _mul_with_hmat_blocks(self, other, out_shape, out_parent_index):
        """multiplication when self and other have blocks
        
        :type other: HMat
        """
        if self.column_sequence() != other.row_sequence():
            raise ValueError('structures are not aligned. '
                             '{0} != {1}'.format(self.column_sequence(), other.row_sequence()))
        out_blocks = []
        #  formula from Howard Eves: Theorem 1.9.6
        for i in list(set([b[0] for b in self.block_structure()])):
            for j in list(set([b[1] for b in other.block_structure()])):
                out_block = None
                for k in list(set([l[1] for l in self.block_structure()])):
                    if out_block is None:
                        out_block = self[i, k] * other[k, j]
                    else:
                        out_block = out_block + self[i, k] * other[k, j]
                out_blocks.append(out_block)
        return HMat(blocks=out_blocks, shape=out_shape, parent_index=out_parent_index)

    def _mul_with_scalar(self, other):
        """multiplication with integer
        
        :type other: Number.number
        """
        out = HMat(shape=self.shape, parent_index=self.parent_index)
        if self.content is not None:
            out.content = self.content * other
            return out
        else:
            out.blocks = [block * other for block in self.blocks]
            return out

    def block_structure(self):
        """return the block structure of self
        
        produce a dict with parent_index: block paris

        :rtype: dict
        :returns: dict with index: HMatrix pairs
        """
        structure = {block.parent_index: block.shape for block in self.blocks}
        return structure

    def column_sequence(self):
        """Return the sequence of column groups
        as in thm. 1.9.4. in :cite:`eves1966elementary`

        only for consistent matrices

        :return: list of column-sizes in first row
        :rtype: list(int)
        :raises: :class:`StructureWarning` if the matrix is not consistent
        """
        if not self.is_consistent():
            raise StructureWarning('Warning, block structure is not consistent! Results may be wrong')
        pre_sort = sorted(self.block_structure(), key=lambda item: item[1])
        sorted_indices = sorted(pre_sort)
        if not sorted_indices:
            return [self.shape[1]]
        start_col = self.parent_index[1]
        current_col = start_col
        max_cols = start_col + self.shape[1]
        col_seq = []
        for index in sorted_indices:
            rows, cols = self.block_structure()[index]
            current_col += cols
            col_seq.append(cols)
            if current_col == max_cols:  # end of column, return list
                return col_seq

    def row_sequence(self):
        """Return the sequence of row groups
        as in thm. 1.9.4. in :cite:`eves1966elementary`

        defined only for consistent matrices

        :return: list of row-sizes in first column
        :rtype: list(int)
        :raises: :class:`StructureWarning` if the matrix is not consistent
        """
        if not self.is_consistent():
            raise StructureWarning('Warning, block structure is not consistent! Results may be wrong')
        pre_sort = sorted(self.block_structure())
        sorted_indices = sorted(pre_sort, key=lambda item: item[1])
        if not sorted_indices:
            return [self.shape[0]]
        start_row = self.parent_index[0]
        current_row = start_row
        max_rows = start_row + self.shape[0]
        row_seq = []
        for index in sorted_indices:
            rows, cols = self.block_structure()[index]
            current_row += rows
            row_seq.append(rows)
            if current_row == max_rows:  # end of row, return list
                return row_seq

    def is_consistent(self):
        """check if the blocks are aligned, 
        i.e. we have consistent rows and columns as in def. 1.9.5 in :cite:`eves1966elementary`

        :return: True on consistency, False otherwise
        :rtype: bool
        """
        if self.block_structure() == {}:  # if we have no blocks, we just have to check the shape
            return self.content.shape == self.shape
        pre_sort = sorted(self.block_structure(), key=lambda item: item[1])
        sorted_indices = sorted(pre_sort)
        start_row, start_col = (0, 0)
        current_row = start_row
        current_col = start_col
        max_rows = start_row + self.shape[0]
        max_cols = start_col + self.shape[1]
        col_rows = 0  # to keep track of the height of each block
        col_seq = []  # sequence of sub-column lengths to compare
        current_col_seq = []
        for index in sorted_indices:
            # iterate over the index list to check each block column by column
            rows, cols = self.block_structure()[index]
            if index != (current_row, current_col):  # starting point of block is not where it should be
                return False
            current_col += cols
            current_col_seq.append(cols)
            if col_rows == 0:  # first block in a column
                col_rows = rows
            if rows != col_rows:  # this block has different height than the others in this column
                return False
            if current_col == max_cols:  # end of column, check against previous and go to next column
                if not col_seq:  # first column, so store for comparison
                    col_seq = current_col_seq
                if col_seq != current_col_seq:  # this column has a different partition than the previous
                    return False
                current_col = start_col
                current_row += col_rows
                col_rows = 0
                current_col_seq = []
        if current_row == max_rows:  # end of row, all fine
            return True
        # if we get to here, the shape of self is exceeded by its blocks in at least one direction
        return False

    def restructure(self, structure):
        """Restructure self into blocks to match structure
        
        :param structure: a dict as returned by HMat.block_structure
        :type structure: dict(index: size)
        :returns: HMat that has the specified structure
        :rtype: HMat
        :raises: :class:`NotImplementedError` if self has blocks or if self contains unknown content
        """
        if self.blocks != ():
            raise NotImplementedError('Only implemented for hmat without blocks')
        out_blocks = []
        for index in structure:
            start_x = index[0]
            start_y = index[1]
            end_x = start_x + structure[index][0]
            end_y = start_y + structure[index][1]
            if isinstance(self.content, RMat):
                out_blocks.append(HMat(content=self.split_rmat(start_x, start_y, end_x, end_y),
                                       shape=structure[index], parent_index=index))
            elif isinstance(self.content, numpy.matrix):
                out_blocks.append(HMat(content=self.split_hmat(start_x, start_y, end_x, end_y),
                                       shape=structure[index], parent_index=index))
            else:
                raise NotImplementedError('Illegal structure found in restructure')
        return HMat(blocks=out_blocks, shape=self.shape, parent_index=self.parent_index)

    def split_rmat(self, start_x, start_y, end_x, end_y):
        """Fetch the block specified by indices from a rmat
        
        .. admonition:: Example
            
            .. code:: python
        
                r = RMat(numpy.matrix([[2], [2], [2]]),
                         numpy.matrix([[3], [3], [3]]))
                h = HMat(content=r, shape=(3, 3), parent_index=(0, 0))
                res = h.split_rmat(0, 2, 1, 3)
                res
                <RMat with left_mat: matrix([[2]]), right_mat: matrix([[3]]) and max_rank: None> 
        
        :param start_x: vertical start index
        :type start_x: int
        
        :param start_y: horizontal start index
        :type start_y: int
        :param end_x: vertical end index
        :type end_x: int
        :param end_y: horizontal end index
        :type end_y: int
        :returns: rmat with corresponding block
        :rtype: RMat
        """
        return self.content.split(start_x, start_y, end_x, end_y)

    def split_hmat(self, start_x, start_y, end_x, end_y):
        """Fetch the block specified by indices from a numpy.matrix
        
        :param start_x: vertical start index
        :type start_x: int
        :param start_y: horizontal start index
        :type start_y: int
        :param end_x: vertical end index
        :type end_x: int
        :param end_y: horizontal end index
        :type end_y: int
        :returns: numpy.matrix with corresponding block
        :rtype: numpy.matrix
        """
        return self.content[start_x: end_x, start_y: end_y]

    def to_matrix(self):
        """Full matrix representation

        :return: full matrix
        :rtype: numpy.matrix
        """
        if self.blocks:  # The matrix has children so fill recursive
            out_mat = numpy.matrix(numpy.zeros(self.shape))
            for block in self.blocks:
                # determine the position of the current block
                vertical_start = block.parent_index[0]
                vertical_end = vertical_start + block.shape[0]
                horizontal_start = block.parent_index[1]
                horizontal_end = horizontal_start + block.shape[1]

                # fill the block with recursive call
                out_mat[vertical_start:vertical_end, horizontal_start:horizontal_end] = block.to_matrix()
            return out_mat
        elif isinstance(self.content, RMat):  # We have an RMat in content, so return its full representation
            return self.content.to_matrix()
        else:  # We have regular content, so we return it
            return self.content

    def transpose(self):
        """Return transposed copy of self

        :rtype: RMat
        """
        out_shape = (self.shape[1], self.shape[0])
        if self.blocks == ():
            out_content = self.content.transpose()
            out_blocks = ()
        else:
            out_content = None
            out_blocks = []
            for block in self.blocks:
                out_block = block.transpose()
                out_block.parent_index = (block.parent_index[1], block.parent_index[0])
                out_blocks.append(out_block)
        return HMat(content=out_content, blocks=out_blocks, shape=out_shape, parent_index=self.parent_index)

    def zero(self, make_copy=True):
        """Return a copy of self with zero entries
        
        i.e. an HMat instance with the same structure, but all content is replaced by numpy.zeros
        
        .. note::
        
            If make_copy is True, a deepcopy is made at the beginning. This can be very slow for large matrices.
        
        :param make_copy: make a copy of the matrix to not change the original (Default True)
        :type make_copy: bool
        :return: zero matrix
        :rtype: HMat
        """
        if self.content is not None:
            return HMat(content=numpy.matrix(numpy.zeros(self.shape)), shape=self.shape, parent_index=self.parent_index)
        else:
            if make_copy:
                out = copy.deepcopy(self)
            else:
                out = self
            structure = out.block_structure()
            for index in structure:
                out[index] = out[index].zero()
            return out

    def inv(self, make_copy=True):
        """Invert the matrix according to chapter 7.5.1 in :cite:`hackbusch2015hierarchical`
         
        .. note::
        
            If make_copy is True, a deepcopy is made at the beginning. This can be very slow for large matrices.
        
        :param make_copy: make a copy of the matrix to not change the original (Default True)
        :type make_copy: bool
        :return: approximation of inverse
        :rtype: HMat
        :raises: :class:`numpy.LinAlgError` if matrix is singular
        """
        if isinstance(self.content, numpy.matrix):
            return HMat(content=numpy.linalg.inv(self.content), shape=self.shape, parent_index=self.parent_index)
        if make_copy:
            self_copy = copy.deepcopy(self)
        else:
            self_copy = self
        out_matrix = self.zero()
        row_sequence = self.row_sequence()
        col_sequence = self.column_sequence()
        bls = row_sequence[0]  # block size
        block_check = [bls != c for c in row_sequence]
        if row_sequence != col_sequence or any(block_check):
            raise ValueError('not all squares')
        rows = len(row_sequence)
        for l in xrange(rows):
            out_matrix[l * bls,  l * bls] = self_copy[l * bls, l * bls].inv()
            for j in xrange(l):
                out_matrix[l * bls, j * bls] = out_matrix[l * bls, l * bls] * out_matrix[l * bls, j * bls]
            for j in xrange(l + 1, rows):
                self_copy[l * bls, j * bls] = out_matrix[l * bls, l * bls] * self_copy[l * bls, j * bls]
            for i in xrange(l + 1, rows):
                for j in xrange(l+1):
                    out_matrix[i * bls,
                               j * bls] = out_matrix[i * bls,
                                                     j * bls] - self_copy[i * bls,
                                                                          l * bls] * out_matrix[l * bls,
                                                                                                j * bls]
                for j in xrange(l+1, rows):
                    self_copy[i * bls,
                              j * bls] = self_copy[i * bls,
                                                   j * bls] - self_copy[i * bls,
                                                                        l * bls] * self_copy[l * bls,
                                                                                             j * bls]
        for l in xrange(rows - 1, -1, -1):
            for i in xrange(l - 1, -1, -1):
                for j in xrange(rows):
                    out_matrix[i * bls,
                               j * bls] = out_matrix[i * bls,
                                                     j * bls] - self_copy[i * bls,
                                                                          l * bls] * out_matrix[l * bls,
                                                                                                j * bls]
        return out_matrix


def build_hmatrix(block_cluster_tree=None, generate_rmat_function=None, generate_full_matrix_function=None):
    """Factory to build an hierarchical matrix
    
    Takes a block cluster tree and generating functions for full matrices and low rank matrices respectively
    
    The generating functions take a block cluster tree as input and return a RMat or numpy.matrix respectively

    :param block_cluster_tree: block cluster tree giving the structure
    :type block_cluster_tree: BlockClusterTree
    :param generate_rmat_function: function taking an admissible block cluster tree and returning a rank-k matrix
    :param generate_full_matrix_function: function taking an inadmissible block cluster tree and returning
        a numpy.matrix
    :return: hmatrix
    :rtype: HMat
    """
    root = HMat(blocks=[], shape=tuple(block_cluster_tree.shape()), parent_index=(0, 0))
    recursion_build_hmatrix(root, block_cluster_tree, generate_rmat_function, generate_full_matrix_function)
    return root


def recursion_build_hmatrix(current_hmat, block_cluster_tree, generate_rmat, generate_full_mat):
    """Recursion to :func:`build_hmatrix`
    """
    if block_cluster_tree.admissible:
        # admissible level found, so fill content with rank-k matrix and stop
        current_hmat.content = generate_rmat(block_cluster_tree)
        current_hmat.blocks = ()
    elif not block_cluster_tree.sons:
        # no sons and not admissible, so fill content with full matrix and stop
        current_hmat.content = generate_full_mat(block_cluster_tree)
        current_hmat.blocks = ()
    else:
        # recursion: generate new hmatrix for every son in block cluster tree
        x_parent, y_parent = block_cluster_tree.plot_info
        for son in block_cluster_tree.sons:
            x_current, y_current = son.plot_info
            parent_index = (x_current - x_parent, y_current - y_parent)
            new_hmat = HMat(blocks=[], shape=son.shape(), parent_index=parent_index)
            current_hmat.blocks.append(new_hmat)
            recursion_build_hmatrix(new_hmat, son, generate_rmat, generate_full_mat)


class StructureWarning(Warning):
    """Special Warning used in this module to indicate bad block structure"""
    pass



"""splitable.py: :class:`Splitable` (Interface) and iterator and different strategies

- :class:`RegularCuboid`

    Starting from a minimal cuboid that get's split in half on every level, points are distributed according to
    their geometric location

- :class:`MinimalCuboid`

    Same as RegularCuboid, but after every split, the cuboids are reduced to minimal size

- :class:`Balanced`

    Distributes the points evenly on every level, depends solely on the order of the initial list
"""
import numpy

from HierMat.cluster import Cluster
from HierMat.cuboid import Cuboid


class Splitable(object):
    """Interface to the different strategies that can be used to restructure a cluster in two or more.

    .. admonition:: Methods that need to be implemented by subclasses

        - __eq__: check for equality against other Splitable
        - split: restructure the object in two or more, return new instances
        - diameter: return the diameter of the object
        - distance: return the distance to other object

    """
    cluster = None

    def __iter__(self):
        return SplitableIterator(self)

    def __repr__(self):
        return '<Splitable with cluster {0}>'.format(self.cluster)

    def __getitem__(self, item):
        return self.cluster[item]

    def __len__(self):
        """Return the length of the cluster"""
        return len(self.cluster)

    def __eq__(self, other):
        raise NotImplementedError()

    def __ne__(self, other):
        return not self == other

    def get_index(self, item):
        """Get index from cluster

        :param item: index to get
        :type item: int
        :return: index
        :rtype: int
        """
        return self.cluster.get_index(item)

    def get_grid_item(self, item):
        """Get grid item from cluster

        :param item: index of item to get
        :type item: int
        """
        return self.cluster.get_grid_item(item)

    def get_grid_item_support(self, item):
        """Return supports of item from grid

        :param item: point
        :type item: tuple(float)
        """
        return self.cluster.get_grid_item_support(item)

    def get_grid_item_support_by_index(self, item):
        """Return supports of item from grid

        :param item: index
        :type item: int
        """
        return self.cluster.get_grid_item_support_by_index(item)

    def get_patch_coordinates(self):
        """Return min and max out of cluster indices

        :return: min and max
        :rtype: tuple(int, int)
        """
        return self.cluster.get_patch_coordinates()

    def diameter(self):
        raise NotImplementedError()

    def distance(self, other):
        raise NotImplementedError()

    def split(self):
        raise NotImplementedError()


class SplitableIterator(object):
    """Iterator to the Splitable implementations."""

    def __init__(self, obj):
        self.obj = obj
        self.counter = 0

    def __iter__(self):
        return self

    def next(self):
        if self.counter >= len(self.obj):
            raise StopIteration
        else:
            self.counter += 1
            return self.obj[self.counter - 1]


class RegularCuboid(Splitable):
    """Method of regular cuboids

    If no cuboid is given, a minimal cuboid is built around initial list of indices. Split is then
    implemented by splitting the surrounding cuboid in half along longest axis and distributing indices according to the
    cuboid they belong to.

    Gives a binary tree
    """

    def __init__(self, cluster, cuboid=None):
        """Build a RegularCuboid

        Args:
            cluster: Cluster instance
            cuboid: Cuboid instance surrounding the cluster. (optional)
        """
        self.cluster = cluster
        self.cuboid = cuboid if cuboid else minimal_cuboid(cluster)

    def __repr__(self):
        return "<RegularCuboid with cluster {0} and cuboid {1}>".format(self.cluster, self.cuboid)

    def __eq__(self, other):
        """:type other: RegularCuboid
        :rtype: bool 
        """
        return self.cluster == other.cluster and self.cuboid == other.cuboid

    def split(self):
        """Split the cuboid and distribute items in cluster according to the cuboid they belong to

        :return: list of smaller regular cuboids 
        :rtype: list(RegularCuboid)
        """
        left_cuboid, right_cuboid = self.cuboid.split()
        left_indices = []
        right_indices = []
        for index in self.cluster.indices:
            if self.cluster.grid.points[index] in left_cuboid:
                left_indices.append(index)
            else:
                right_indices.append(index)
        if len(left_indices) > 0:
            left_cluster = Cluster(self.cluster.grid, left_indices)
            left_rc = RegularCuboid(left_cluster, left_cuboid)
        else:
            left_rc = None
        if len(right_indices) > 0:
            right_cluster = Cluster(self.cluster.grid, right_indices)
            right_rc = RegularCuboid(right_cluster, right_cuboid)
        else:
            right_rc = None
        if not left_rc and right_rc:
            return [right_rc]
        elif not right_rc and left_rc:
            return [left_rc]
        else:
            return [left_rc, right_rc]

    def diameter(self):
        """Return the diameter of the cuboid

        Diameter of the surrounding cuboid

        :return: diameter
        :rtype: float
        """
        return self.cuboid.diameter()

    def distance(self, other):
        """Return the distance between own cuboid and other cuboid

        :param other: other regular cuboid
        :type other: RegularCuboid
        :return: distance
        :rtype: float
        """
        return self.cuboid.distance(other.cuboid)


class MinimalCuboid(RegularCuboid):
    """Method of minimal cuboids

    Split is implemented by splitting the surrounding cuboid in half along longest axis and distributing indices 
    according to the cuboid they belong to. Afterwards all resulting cuboids are shrunk to minimal.

    Gives a binary tree
    """

    def __init__(self, cluster):
        """Build a MinimalCuboid

        :param cluster: cluster to surround
        :type cluster: Cluster
        """
        super(MinimalCuboid, self).__init__(cluster)

    def __repr__(self):
        return "<MinimalCuboid with cluster {0} and cuboid {1}>".format(self.cluster, self.cuboid)

    def split(self):
        """Split the cuboid and distribute items in cluster according to the cuboid they belong to. Reduce every restructure
        to minimal size

        :return: list of minimal cuboids of smaller size
        :rtype: list(MinimalCuboid)
        """
        splits = super(MinimalCuboid, self).split()
        outs = [MinimalCuboid(split.cluster) for split in splits]
        return outs


class Balanced(Splitable):
    """Balanced strategy, where a cluster is always restructure in half only according to its size, without regarding the
    geometry
    
    Gives a binary tree
    """

    def __init__(self, cluster):
        """
        
        :param cluster: a cluster instance
        :type cluster: Cluster
        """
        self.cluster = cluster

    def __eq__(self, other):
        return self.cluster == other.cluster

    def __len__(self):
        return len(self.cluster)

    def __getitem__(self, item):
        return self.cluster[item]

    def __repr__(self):
        return "<Balanced with cluster {0}>".format(self.cluster)

    def get_index(self, item):
        """Get index from cluster

        :param item: index to get
        :type item: int
        :return: index
        :rtype: int
        """
        return self.cluster.get_index(item)

    def get_grid_item(self, item):
        """Get grid item from cluster

        :param item: index of item to get
        :type item: int
        """
        return self.cluster.get_grid_item(item)

    def get_grid_item_support(self, item):
        """Return supports of item from grid

        :param item: point
        :type item: tuple(float)
        """
        return self.cluster.get_grid_item_support(item)

    def get_grid_item_support_by_index(self, item):
        """Return supports of item from grid

        :param item: index
        :type item: int
        """
        return self.cluster.get_grid_item_support_by_index(item)

    def get_patch_coordinates(self):
        """Return min and max out of indices

        :return: min and max
        :rtype: tuple(int, int)
        """
        return self.cluster.get_patch_coordinates()

    def diameter(self):
        """Return the diameter

        Diameter of the contained cluster

        :return: diameter
        :rtype: float
        """
        return self.cluster.diameter()

    def distance(self, other):
        """Distance to other balanced
        
        :param other: other balanced
        :type other: Balanced
        :return: distance
        :rtype: float
        """
        return self.cluster.distance(other.cluster)

    def split(self):
        """Split in two
        
        distribute the first (smaller) half to the left and the rest to the right
        
        return itself if it has only one point
        
        :rtype: list(Balanced)
        """
        split = int(len(self)/2)
        if split >= 1:
            left_cluster = Cluster(self.cluster.grid, self.cluster.indices[:split])
            right_cluster = Cluster(self.cluster.grid, self.cluster.indices[split:])
            return [Balanced(left_cluster), Balanced(right_cluster)]
        else:
            return self


def minimal_cuboid(cluster):
    """Build minimal cuboid

    Build minimal cuboid around cluster that is parallel to the axis in Cartesian coordinates

    :param cluster: cluster to build cuboid around
    :type cluster: Cluster
    :return: minimal cuboid
    :rtype: Cuboid
    """
    points = cluster.grid.points
    low_corner = numpy.array(points[0], float, ndmin=1)
    high_corner = numpy.array(points[0], float, ndmin=1)
    for p in points:
        p = numpy.array(p, float, ndmin=1)
        lowers = (p < low_corner).nonzero()
        for l in lowers:
            low_corner[l] = p[l]
        highers = (p > high_corner).nonzero()
        for h in highers:
            high_corner[h] = p[h]
    return Cuboid(low_corner, high_corner)


"""This is an implementation of the one-dimensional model example described in :cite:`thesis`.
It shows a typical use-case of hierarchical matrices:
    
.. admonition:: Integral Equation
    
    .. math::
        
        \int_0^1 \frac{1}{2\pi}\log |x-y| u(y) dy = g
    
    for :math:`x \in [0,1]`.

    After Galerkin discretization, we end up with a linear system

    .. math::

        \mathbf{A}^{\\text{Gal}} \cdot \\alpha = \mathbf{g}

    where

    .. math::

        \mathbf{A}^{\\text{Gal}}_{t,\\tau}:= \int_{t}\int_{\\tau} \log |x-y| dy dx

    To determine admissible blocks we start by building the geometric objects:

    .. code-block:: python

        midpoints = [((i + 0.5)/n,) for i in xrange(n)]
        intervals = {p: (p[0] - 0.5/n, p[0] + 0.5/n) for p in midpoints}
        grid = HierMat.Grid(points=midpoints, supports=intervals)
        cluster = HierMat.Cluster(grid=grid)
        unit_cuboid = HierMat.Cuboid([0], [1])
        strategy = HierMat.RegularCuboid(cluster=cluster, cuboid=unit_cuboid)
        cluster_tree = HierMat.build_cluster_tree(splitable=strategy, max_leaf_size=n_min)
        block_cluster_tree = HierMat.build_block_cluster_tree(left_cluster_tree=cluster_tree,
                                                          right_cluster_tree=cluster_tree,
                                                          admissible_function=HierMat.admissible
                                                          )

    With the structure established, we can produce the hierarchical matrix:

    .. code-block:: python

        hmat = HierMat.build_hmatrix(block_cluster_tree=block_cluster_tree,
                                 generate_rmat_function=lambda bct: galerkin_1d_rank_k(bct, max_rank),
                                 generate_full_matrix_function=galerkin_1d_full
                                 )
"""
import numpy
import scipy.integrate as integrate

import HierMat

import os
import math


def model_1d(n=2 ** 3, max_rank=1, n_min=1):
    """This is an implementation of the one-dimensional model example described in :cite:`thesis`.

    :param n: number of discretization points
    :type n: int
    :param max_rank: max rank of the low-rank approximation
    :type max_rank: int
    :param n_min: minimal leaf size for cluster trees
    :type n_min: int
    :return: 0 (just for testing)
    """
    midpoints = [((i + 0.5)/n,) for i in xrange(n)]
    intervals = {p: (p[0] - 0.5/n, p[0] + 0.5/n) for p in midpoints}
    grid = HierMat.Grid(points=midpoints, supports=intervals)
    cluster = HierMat.Cluster(grid=grid)
    unit_cuboid = HierMat.Cuboid([0], [1])
    strategy = HierMat.RegularCuboid(cluster=cluster, cuboid=unit_cuboid)
    cluster_tree = HierMat.build_cluster_tree(splitable=strategy, max_leaf_size=n_min)
    HierMat.export(cluster_tree, form='dot', out_file='galerkin_1d_ct.dot')
    os.system('dot -Tpng galerkin_1d_ct.dot > model_1d-ct.png')
    os.system('dot -Tsvg galerkin_1d_ct.dot > model_1d-ct.svg')
    block_cluster_tree = HierMat.build_block_cluster_tree(left_cluster_tree=cluster_tree,
                                                          right_cluster_tree=cluster_tree,
                                                          admissible_function=HierMat.admissible
                                                          )
    HierMat.plot(block_cluster_tree, filename='model_1d-bct.png')
    hmat = HierMat.build_hmatrix(block_cluster_tree=block_cluster_tree,
                                 generate_rmat_function=lambda bct: galerkin_1d_rank_k(bct, max_rank),
                                 generate_full_matrix_function=galerkin_1d_full
                                 )
    hmat_full = hmat.to_matrix()
    galerkin_full = galerkin_1d_full(block_cluster_tree)
    HierMat.export(hmat, form='bin', out_file='hmat.bin')
    numpy.savetxt('hmat_full.txt', hmat_full)
    numpy.savetxt('gallmat_full.txt', galerkin_full)
    return 0


def kerlog(x):
    """kerlog function as in :cite:`thesis`.

    .. math::

        kerlog(x):= x^2 \left( \log(\Vert x \Vert) - \\frac{1}{2} \\right)

    :param x: real number
    :type x: float
    :return: :math:`x^2 ( \log(\Vert x \Vert) - \\frac{1}{2} )`
    :rtype: float
    """
    out = x**2 * (numpy.log(abs(x))-0.5)
    if math.isnan(out):
        return 0
    else:
        return out


def ker(x, y):
    """Kernel to integrate

    .. math::

        ker(x, y):= \log\left(\Vert x - y \Vert\\right)

    :param x: real number
    :type x: float
    :param y: real number
    :type y: float
    :return: :math:`\log\left(\Vert x - y \Vert\\right)`
    :rtype: float
    """
    return numpy.log(abs(x - y)) / (2 * numpy.pi)


def galerkin_1d_rank_k(block_cluster_tree, max_rank):
    """Low-rank approximation of the kernel

    .. math::

        R: &= A \cdot B^T

        A_{\\tau, k}: &= \int_\\tau \log \Vert x-y_k\Vert dx

        B_{\\tau, k}: &= \int_\\tau \mathcal{L}_k(y) dy

    
    :param block_cluster_tree: admissible block cluster tree
    :type block_cluster_tree: HierMat.BlockClusterTree
    :param max_rank: separation rank
    :type max_rank: int
    :return: 
    """
    # initialize output
    x_length, y_length = block_cluster_tree.shape()
    left_matrix = numpy.matrix(numpy.zeros((x_length, max_rank)))
    right_matrix = numpy.matrix(numpy.zeros((y_length, max_rank)))
    # determine the y-interval
    y_low = block_cluster_tree.right_clustertree.get_grid_item_support_by_index(0)[0]
    y_high = block_cluster_tree.right_clustertree.get_grid_item_support_by_index(-1)[1]
    # build Chebyshev nodes
    y_nodes = get_chebyshev_interpol_points(max_rank, y_low, y_high)
    # build left_hand matrix
    y_count = 0
    for y_k in y_nodes:
        for i in xrange(x_length):
            lower, upper = block_cluster_tree.left_clustertree.get_grid_item_support_by_index(i)
            left_matrix[i, y_count] = integrate.quad(lambda x: ker(x, y_k), lower, upper)[0]
        for j in xrange(y_length):
            lower, upper = block_cluster_tree.right_clustertree.get_grid_item_support_by_index(j)

            def lagrange(y):
                res = 1.0
                for y_i in y_nodes:
                    if y_i != y_k:
                        res *= (y - y_i) / (y_k - y_i)
                return res
            right_matrix[j, y_count] = integrate.quad(lagrange, lower, upper)[0]

    return HierMat.RMat(left_mat=left_matrix, right_mat=right_matrix, max_rank=max_rank)


def galerkin_1d_full(block_cluster_tree):
    """Exact calculation of the integral

    .. math::

        A_{i,j}=A_{\\tau,t}^{Gal}=\int_t\int_\\tau\log\Vert x-y\Vert \;dydx
    
    :param block_cluster_tree: inadmissible block cluster tree
    :type block_cluster_tree: HierMat.BlockClusterTree
    :return: matrix with same shape as block_cluster_tree.shape()
    :rtype: numpy.matrix
    """
    x_length, y_length = block_cluster_tree.shape()
    out_matrix = numpy.matrix(numpy.zeros((x_length, y_length)))
    for i in xrange(x_length):
        for j in xrange(y_length):
            c, d = block_cluster_tree.left_clustertree.get_grid_item_support_by_index(i)
            a, b = block_cluster_tree.right_clustertree.get_grid_item_support_by_index(j)
            out_matrix[i, j] = 1.0 / (2 * numpy.pi) * (-kerlog(d - b) + kerlog(c - b) + kerlog(d - a) - kerlog(c - a)) \
                               / 2 + (a - b) * (d - c)
    return out_matrix


def get_chebyshev_interpol_points(points, lower=0, upper=1):
    """Get Chebyshev interpolation points on interval :math:`(a, b)`.

    .. math::

        \\frac{1}{2} (a + b) + \\frac{1}{2} (b - a) \cos\left(\\frac{2k+1}{2n}\pi\\right)

    for :math:`k=0, ..., \\text{points} - 1`

    :param points: number of points
    :type points: int
    :param lower: :math:`a`, (default 0)
    :type lower: float
    :param upper: :math:`b`, (default 0)
    :type upper: float
    :return: :math:`\\frac{1}{2} (a + b) + \\frac{1}{2} (b - a) \cos\left(\\frac{2k+1}{2n}\pi\\right)`
    :rtype: list(floats)
    """
    return [(lower + upper + (upper - lower) * numpy.cos((2 * k + 1) * numpy.pi / 2 * points)) / 2
            for k in xrange(points)]


if __name__ == '__main__':
    model_1d(n=2**3, max_rank=2, n_min=1)



"""Examples to the HierMat package

In this module are example scripts to demonstrate the usage of the package

model_1d module
------------------

.. automodule:: HierMat.examples.model_1d
    :members:

"""


"""cluster_tree.py: :class:`ClusterTree`, :func:`build_cluster_tree`, :func:`recursion_build_cluster_tree`
"""


class ClusterTree(object):
    """Tree structure built according to the :class:`Splitable` in use

    Splits are done until max_leaf_size is reached
    """
    def __init__(self, content, sons=None, max_leaf_size=1, level=0):
        """"""
        self.level = level
        self.content = content
        self.sons = sons if sons else []
        self.max_leaf_size = max_leaf_size

    def __repr__(self):
        optional_string = " with children {0}".format(self.sons) if self.sons else " without children"
        return "<ClusterTree at level {0}{1}>".format(str(self.level), optional_string)

    def __len__(self):
        return len(self.content)

    def __str__(self):
        """give str representation of self."""
        cont_str = ",".join([str(p) for p in self.content])
        out_str = "ClusterTree at level {0} with content:\n{1}".format(self.level, cont_str)
        return out_str

    def _plot_str(self):
        """string for plots"""
        cont_str = ''
        for p in self.content:
            cont_str += '['
            cont_str += ",".join(["{0:.2f}".format(i) for i in p])
            cont_str += "] "
        cont_str.rstrip()
        return cont_str

    def __eq__(self, other):
        """Test for equality
        :param other: other cluster tree
        :type other: ClusterTree
        """
        return (self.level == other.level and self.content == other.content
                and self.sons == other.sons and self.max_leaf_size == other.max_leaf_size)

    def __ne__(self, other):
        """Test for inequality
        :param other: other cluster tree
        :type other: ClusterTree
        """
        return not self == other

    def __getitem__(self, item):
        return self.content[item]

    def get_index(self, item):
        """Get index from content

        :param item: index to get
        :type item: int
        :return: index
        :rtype: int
        """
        return self.content.get_index(item)

    def get_grid_item(self, item):
        """Get grid item from content

        :param item: index of item to get
        :type item: int
        """
        return self.content.get_grid_item(item)

    def get_grid_item_support_by_index(self, item):
        """Return supports of i-th item from grid

        :param item: index
        :type item: int
        """
        return self.content.get_grid_item_support_by_index(item)

    def get_grid_item_support(self, item):
        """Return supports ofitem from grid

        :param item: point
        :type item: tuple(float)
        """
        return self.content.get_grid_item_support(item)

    def get_patch_coordinates(self):
        """Return min and max out of indices

        :return: min and max
        :rtype: tuple(int, int)
        """
        return self.content.get_patch_coordinates()

    def to_list(self):
        """Give list representation for export

        Return a list containing the object and a list with its sons

        .. hint::

            **recursive function**
        """
        if self.sons:
            return [self, [son.to_list() for son in self.sons]]
        else:
            return [self, []]

    def to_xml(self):
        """Return a string for xml representation"""
        out_list = self.to_list()
        return self._to_xml(out_list)

    @staticmethod
    def _to_xml(lst, out_string=''):
        if len(lst[1]):
            value_string = str(lst[0])
            display_string = str(len(lst[1]))
        else:
            value_string = str(lst[0])
            display_string = str(lst[0])
        out_string += '<node value="{0}">{1}\n'.format(value_string, display_string)
        if len(lst) > 1 and type(lst[1]) is list:
            for item in lst[1]:
                out_string = ClusterTree._to_xml(item, out_string)
        out_string += "</node>\n"
        return out_string

    def to_dot(self):
        """Return a string for .dot representation"""
        out_list = self.to_list()
        return self._to_dot(out_list)

    @staticmethod
    def _to_dot(lst, out_string=''):
        value_string = str(lst[0])
        label_string = len(lst[0])
        for item in lst[1]:
            item_string = str(item[0])
            out_string += '''"{0}" -- "{1}";
            "{0}"[label="{2}",color="#cccccc",style="filled",shape="box"];\n'''.format(
                value_string, item_string, label_string)
            out_string = ClusterTree._to_dot(item, out_string)
        if not lst[1]:
            try:
                content_list = ['{:.2f}'.format(p) for p in lst[0].content]
            except ValueError:

                content_list = [str(['{:.2f}'.format(p) for p in item]) for item in lst[0].content]
            label_string = ', '.join(content_list)
            out_string += '"{0}"[label="{1}",color="#cccccc",style="filled",shape="box"];\n'.format(value_string,
                                                                                                    label_string)
        return out_string

    def depth(self, root_level=None):
        """Get depth of the tree

        :param root_level: internal use.
        :type root_level: int
        :return: depth
        :rtype: int

        .. hint::

            **recursive function**
        """
        if root_level is None:
            root_level = self.level
        if not self.sons:
            return self.level - root_level
        else:
            return max([son.depth(root_level) for son in self.sons])

    def diameter(self):
        """Return the diameter of content

        :return: diameter
        :rtype: float
        """
        return self.content.diameter()

    def distance(self, other):
        """Return distance to other

        :param other: other cluster tree
        :type other: ClusterTree
        :return: distance
        :rtype: float
        """
        return self.content.distance(other.content)


def build_cluster_tree(splitable, max_leaf_size=1, start_level=0):
    """Factory for building a cluster tree

    :param splitable: strategy that decides the structure
    :type splitable: Splitable
    :param max_leaf_size: cluster size at which recursion stops
    :type max_leaf_size: int
    :param start_level: counter to identify levels
    :type start_level: int
    :return: cluster tree
    :rtype: ClusterTree
    """
    root = ClusterTree(splitable, max_leaf_size=max_leaf_size, level=start_level)
    recursion_build_cluster_tree(root)
    return root


def recursion_build_cluster_tree(current_tree):
    """Recursion to :func:`build_cluster_tree`
    """
    if len(current_tree.content) > current_tree.max_leaf_size:
        splits = current_tree.content.split()
        for split in splits:
            new_tree = ClusterTree(content=split,
                                   max_leaf_size=current_tree.max_leaf_size,
                                   level=current_tree.level + 1)
            current_tree.sons.append(new_tree)
            recursion_build_cluster_tree(new_tree)


def admissible(left_clustertree, right_clustertree):
    """Default admissible condition for BlockClusterTree

    True if the smaller diameter of the input is smaller or equal to the distance between the two ClusterTrees

    :param left_clustertree: "Left-side" ClusterTree
    :param right_clustertree: "Right-side" ClusterTree
    :type left_clustertree: ClusterTree
    :type right_clustertree: ClusterTree
    :return: admissible
    :rtype: bool
    """
    diam_min = min(left_clustertree.diameter(), right_clustertree.diameter())
    distance = left_clustertree.distance(right_clustertree)
    return diam_min <= distance



"""cuboid.py: :class:`Cuboid` and :func:`minimal_cuboid`
"""
import numpy


class Cuboid(object):
    """Cuboid that is parallel to the axis in Cartesian coordinates.

    Characterized by two diagonal corners.

    - **Attributes**:

        low_corner: numpy.array with minimal values in each dimension

        high_corner: numpy.array with maximal values in each dimension
    """
    def __init__(self, low_corner, high_corner):
        """Build a cuboid.

        :param low_corner: low corner
        :type low_corner: numpy.array
        :param high_corner: high corner
        :type high_corner: numpy.array
        """
        if not hasattr(low_corner, '__len__'):
            low_corner = (low_corner,)
        if not hasattr(high_corner, '__len__'):
            high_corner = (high_corner,)
        if len(low_corner) != len(high_corner):
            raise ValueError('corners must have same dimension')
        self.low_corner = numpy.array(low_corner, float)
        self.high_corner = numpy.array(high_corner, float)

    def __len__(self):
        """Dimension of the corners
        
        :return: len(high_corner)
        :rtype: int
        """
        return len(self.high_corner)

    def __eq__(self, other):
        """Test for equality

        :param other: other cuboid
        :type other: Cuboid
        :return: equal
        :rtype: bool
        """
        if len(self) != len(other):
            return False
        low_eq = self.low_corner == other.low_corner
        high_eq = self.high_corner == other.high_corner
        return low_eq.all() and high_eq.all()

    def __ne__(self, other):
        """Test for inequality

        :param other: other cuboid
        :type other: Cuboid
        :return: equal
        :rtype: bool
        """
        return not self == other

    def __contains__(self, point):
        """Check if point is inside the cuboid

        True if point is between low_corner and high_corner

        :param point: point of correct dimension
        :type point: tuple(float)

        :return: contained
        :rtype: bool
        """
        lower = self.low_corner <= point
        higher = point <= self.high_corner
        return lower.all() and higher.all()

    def __repr__(self):
        """Representation of Cuboid"""
        return "Cuboid({0},{1})".format(str(self.low_corner), str(self.high_corner))

    def __str__(self):
        """Return string representation"""
        return "Cuboid with:\n\tlow corner: {0},\n\thigh corner{1}.".format(str(self.low_corner),
                                                                            str(self.high_corner))

    def split(self, axis=None):
        """Split the cuboid in half

        If axis is specified, the cuboid is restructure along the given axis, else the maximal axis is chosen.

        :param axis: axis along which to restructure (optional)
        :type axis: int
        :return: cuboid1, cuboid2
        :rtype: tuple(Cuboid, Cuboid)
        """
        if axis:
            index = axis
        else:
            # determine dimension in which to restructure
            index = numpy.argmax(abs(self.high_corner - self.low_corner))
        # determine value at splitting point
        split = (self.high_corner[index] + self.low_corner[index]) / 2
        low_corner1 = numpy.array(self.low_corner)
        low_corner2 = numpy.array(self.low_corner)
        low_corner2[index] = split
        high_corner1 = numpy.array(self.high_corner)
        high_corner2 = numpy.array(self.high_corner)
        high_corner1[index] = split
        return Cuboid(low_corner1, high_corner1), Cuboid(low_corner2, high_corner2)

    def diameter(self):
        """Return the diameter of the cuboid.

        Diameter is the Euclidean distance between low_corner and high_corner.

        :return: diameter
        :rtype: float
        """
        return numpy.linalg.norm(self.high_corner - self.low_corner)

    def distance(self, other):
        """Return distance to other Cuboid.

        Distance is the minimal Euclidean distance between points in self and points in other.

        :param other: other cuboid
        :type other: Cuboid

        :return: distance
        :rtype: float
        """
        dimension = len(self.low_corner)
        distance1 = self.low_corner - other.low_corner
        distance2 = self.low_corner - other.high_corner
        distance3 = self.high_corner - other.low_corner
        distance4 = self.high_corner - other.high_corner
        distance_matrix = numpy.array((distance1, distance2, distance3, distance4))
        checks = abs(numpy.sum(numpy.sign(distance_matrix), 0)) == 4 * numpy.ones(dimension)
        distance_vector = numpy.array(checks, dtype=float)
        min_vector = numpy.amin(abs(distance_matrix), axis=0)
        return numpy.linalg.norm(min_vector * distance_vector)



"""grid.py: :class:`Grid` object and iterator
"""
import numpy


class Grid(object):
    """Discretized grid characterized by points and supports
    
    :param points: list of coordinates
    
        The points of the discretized Grid.
            
    :type points: list[tuple(float)]
    :param supports: dictionary mapping points with their supports
    :type supports: dict{point: support}
    :raise ValueError: if points and supports have different length
    """
    def __init__(self, points, supports):
        """Create a Grid"""
        # check input
        if len(points) != len(supports):
            raise ValueError('points and supports must be of same length')
        # fill instance
        self.points = points
        self.supports = supports

    def __len__(self):
        """Return number of points

        :return: len(points)
        :rtype: int
        """
        return len(self.points)

    def __getitem__(self, item):
        """Return point at item

        :param item: index to return
        :type item: int
        """
        return self.points[item]

    def __iter__(self):
        """Iterate trough Grid
        """
        return GridIterator(self)

    def __eq__(self, other):
        """Test for equality

        :param other: other grid
        :type other: Grid
        :return: True on equality
        :rtype: bool
        """
        points_eq = numpy.array_equal(self.points, other.points)
        links_eq = numpy.array_equal(self.supports, other.supports)
        return points_eq and links_eq

    def __ne__(self, other):
        """Test for inequality

        :param other: other grid
        :type other: Grid
        :return: True on inequality
        :rtype: bool
        """
        return not self == other

    def get_point(self, item):
        """Return point at position item

        :param item: index
        :type item: int
        :return: point
        """
        return self.points[item]

    def get_support(self, item):
        """Return support for item

        :param item: point
        :type item: float or tuple(float)
        :return: support
        """
        return self.supports[item]

    def get_support_by_index(self, index):
        """Return support for the i-th item

        :param index: index
        :type index: int
        :return: support
        """
        return self.supports[self.get_point(index)]

    def dim(self):
        """Dimension of the Grid
        
        .. note::
            
            this is the dimension of the first point
             
             if you store points of different dimensions, this will be misleading

        :return: dimension
        :rtype: int
        """
        return len(self[0])


class GridIterator(object):
    """Iterator to Grid object
    """
    def __init__(self, grid):
        self.grid = grid
        self._counter = 0

    def __iter__(self):
        return self

    def next(self):
        if self._counter >= len(self.grid):
            raise StopIteration
        else:
            self._counter += 1
            return self.grid[self._counter - 1]



from block_cluster_tree import BlockClusterTree, build_block_cluster_tree
from cluster import Cluster
from cluster_tree import ClusterTree, build_cluster_tree, admissible
from cuboid import Cuboid
from grid import Grid
from hmat import HMat, build_hmatrix
from rmat import RMat
from splitable import RegularCuboid, MinimalCuboid, Balanced, Splitable
from utils import plot, export, load, divisor_generator
from examples.model_1d import model_1d



"""
HierMat
=======

This library is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this package.  If not, see <http://www.gnu.org/licenses/>.
"""
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HierMat.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup

with open('README.rst', 'r') as readme:
    long_desc = readme.read()

setup(
    name='HierMat',
    version='0.8.0',
    packages=['HierMat', 'HierMat/examples'],
    url='http://hierarchicalmatrices.readthedocs.io/en/latest/index.html',
    download_url='https://github.com/maekke97/HierarchicalMatrices',
    license='GNU GPL v3',
    author='Markus Neumann',
    author_email='markus.neumann@math.uzh.ch',
    requires=['numpy', 'matplotlib', 'scipy'],
    description='Framework for Hierarchical Matrices',
    long_description=long_desc
)
